{% extends "reports/base.html" %}
{% load report_tags %}

{% block content %}
<table>
  <tr>
    <th>Received
    <th>User
    <th>gfx_card
    <th>gfx_drv_ver
    <th>uname_machine
    <th>uname_release
    <th>uname_sysname
    <th>uname_version
{% for report in report_page.object_list %}
{% with json=report.data_json %}
  <tr {% cycle 'class=alt' '' %}>
    <td>{{ report.upload_date|date:"Y-m-d" }}&nbsp;{{ report.upload_date|date:"H:i:s" }}
    <td><a href="{% url userreport.views_private.report_user report.user_id_hash %}"><abbr title="{{ report.user_id_hash }}">{{ report.user_id_hash|slice:"0:8" }}</abbr></a>
    <td>{{ json.gfx_card }}
    <td>{{ json.gfx_drv_ver }}
    <td>{{ json.uname_machine }}
    <td>{{ json.uname_release }}
    <td>{{ json.uname_sysname }}
    <td>{{ json.uname_version }}
{% endwith %}
{% endfor %}
</table>
{% endblock %}