#include <cxxtest/PsTestWrapper.h>

namespace CxxTest 
{
    const char *g_PsArgv0 = NULL;
    const char *g_PsOnlySuite = NULL;
    const char *g_PsOnlyTest = NULL;
    bool g_PsListSuites = false;
}
