/* Copyright (C) 2009 Wildfire Games.
 * This file is part of 0 A.D.
 *
 * 0 A.D. is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * 0 A.D. is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 0 A.D.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef INCLUDED_SCRIPTGLUE
#define INCLUDED_SCRIPTGLUE

#include "ScriptingHost.h"

// referenced by ScriptingHost.cpp
extern JSFunctionSpec ScriptFunctionTable[];
extern JSPropertySpec ScriptGlobalTable[];

// dependencies (moved to header to avoid L4 warnings)
// .. from main.cpp:
extern int fps;
extern void kill_mainloop();
extern CStrW g_CursorName;
extern void StartGame();
extern void EndGame();
// .. other
#if OS_WIN
extern int GetVRAMInfo(int&, int&);
#endif

#endif	// #ifndef INCLUDED_SCRIPTGLUE
