/*
	Copyright (C) 2005-2007 Feeling Software Inc.
	Portions of the code are:
	Copyright (C) 2005-2007 Sony Computer Entertainment America
	
	MIT License: http://www.opensource.org/licenses/mit-license.php
*/

/**
	@file FCDEmitterObject.h
	This file contains the FCDEmitterObject class, and definitions
	for its two children FCDPositionObject and FCDOrienationObject
	and their various definitions
*/

#ifndef _FCD_EMITTER_OBJECT_H_
#define _FCD_EMITTER_OBJECT_H_

#ifndef _FCD_OBJECT_H_
#include "FCDocument/FCDObject.h"
#endif // _FCD_OBJECT_H_
#ifndef _FCD_PARAMETER_ANIMATABLE_H_
#include "FCDocument/FCDParameterAnimatable.h"
#endif // _FCD_PARAMETER_ANIMATABLE_H_

class FCDGeometry;
class FCDEntityReference;
class FUUri;

#endif
